﻿//*************************************************************************************************
//【サンプルプログラム】『Kinectから骨格情報を取得し、右手左手の残像履歴を表示する』
// 人物切り取り表示処理を含む
//【名称】Bitmap_GetSkeletonAfterimage　　【Version】1.02
//【作成者】私設研究所Neo-Tech-Lab.com  上田智章
//【作成日】2012年3月5日　【コメント追記】2012年6月30日
//【注意事項】
//　このプログラムをKinect for Windows SDK Version1.5で動かすには、参照設定の
// Microsoft.Kinectを一旦削除し、プロジェクト(P)-参照の追加(F)で、.NETタブ内
// にあるMicrosoft.Kinectを選択・追加してください。 
//*************************************************************************************************

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Diagnostics;             //【Debug】
using System.Drawing;                 //【Bitmap】【Color】
using System.Drawing.Drawing2D;
using System.Drawing.Imaging;         //【BitmapData】
using System.Runtime.InteropServices; //【Marshal.Copy】
using Microsoft.Kinect;               //【参照設定】プロジェクト(P)⇒参照の追加(F)⇒.NETタブ⇒Microsoft.Kinectを選択

namespace Bitmap_GetSkeletonAfterimage
{
    public partial class Form1 : Form
    {
        //【KINECT関連】
        //　関数内でダイナミックに変数を確保すると動作速度が遅くなるので、ここにまとめて変数を確保しています。
        private KinectSensor KS_KinectSensor;   //【KINECT】
        private ColorImageFrame KS_ColorFrame;  //【KINECTのColorImageFrame】
        private byte[] KS_ColorImageDataXRGB;   //【KINECTのColorImageData】32bit/画素,X8R8G8B8表示形式
        private byte[] KS_ColorImageDataARGB;   //【KINECTのColorImageData】32bit/画素,A8R8G8B8表示形式
        private int KS_ColorFrameWidth;         //【KINECTのColorFrameの横方向画素数】
        private int KS_ColorFrameHeight;        //【KINECTのColorFrameの縦方向画素数】
        private Bitmap KS_bmp1;                 //【ColorImage表示用Bitmapオブジェクト】
        private BitmapData KS_data1;            //【ColorImage用BitmapData】
        private Rectangle KS_rect1;             //【ColorImage表示枠】
        private DepthImageFrame KS_DepthFrame;  //【KINECTのDepthImageFrame】
        private short[] KS_DepthData;           //【KINECTのDepthImageData】16bit/画素(MSB13bit:Depth, LSB3bit:PlayerID) 
        private byte[] KS_DepthImageDataARGB;   //【KINECTのDepthImageData】32bit/画素,A8R8G8B8表示形式
        private int KS_DepthFrameWidth;         //【KINECTのDepthFrameの横方向画素数】
        private int KS_DepthFrameHeight;        //【KINECTのDepthFrameの縦方向画素数】
        private Bitmap KS_bmp2;                 //【DepthImage表示用Bitmapオブジェクト】
        private BitmapData KS_data2;            //【DepthImage用BitmapData】
        private Rectangle KS_rect2;             //【DepthImage表示枠】
        private int KS_PlayerIndexBitmask;      //【DepthデータのPlayer情報】 0000 0000 0000 0111B = 7
        private int KS_PlayerIndexBitmaskWidth; //【DepthデータのPlayer情報】 3 bits
        private int KS_Lmin;                    //【深度測定可能最近距離】default:  800mm, near: 400mm 
        private int KS_Lmax;                    //【深度測定可能最遠距離】default: 4000mm, near:3000mm
        private SkeletonFrame KS_SkeletonFrame; //【KINECTのSkeletonFrame】
        private Skeleton[] KS_SkeletonData;     //【KINECTのSkeleton】Skeleton 最大6名、20Joints/Player
        private bool KS_flagFlip = true;        //【左右反転フラグ】
        private bool KS_flagPlayer = false;     //【Player検出フラグ】
        //【NTL拡張現実センサ関連】---------------------------------------------------------
        private NTL_ColorBar ColorBar = new NTL_ColorBar(); //【デプス値をカラーに変換する部品】
        private string[] BackGroundFileName = { "../../image/MMD.bmp", "../../image/Sky.bmp" };
        private bool NTL_flagMappingInfo = true;
        private ColorImagePoint NTL_Pc1; //【逆マッピング用】
        private ColorImagePoint NTL_Pc2; //　カラーイメージ⇒デプスイメージ
        private ColorImagePoint NTL_Pc3; //
        private float NTL_cDx;           //
        private float NTL_cDy;           //
        private SkeletonData NTL_SkeletonData;
        private const int NTL_default_nFIFO = 120; //【FIFO段数】Default 30FPSで2秒分に相当
        private int NTL_nFIFO = NTL_default_nFIFO; //【FIFO段数】
        private int NTL_wPtr;                      //【FIFO書込みポインタ】
        private int NTL_rPtr;                      //【FIFO読出しポインタ】
        private bool NTL_FIFOfull = false;         //【FIFOフル】
        private bool NTL_FIFOempty = true;         //【FIFO空】
        private NTL_FIFOData[] NTL_FIFO_RH = new NTL_FIFOData[NTL_default_nFIFO];
        private NTL_FIFOData[] NTL_FIFO_LH = new NTL_FIFOData[NTL_default_nFIFO];
        private NTL_FIFOData NTL_RightHand;        //【右手】
        private NTL_FIFOData NTL_LeftHand;         //【左手】

        public struct NTL_FIFOData
        {
            public SkeletonPoint Ps;
            public DepthImagePoint Pd;
            public ColorImagePoint Pc;
            public Color ColorValue;
            public int SensorValue;
        }

        public Form1()
        {
            InitializeComponent();
            //***********************************************************
            //【Kinect検出】【共通】
            //***********************************************************
            //【利用可能なKINECTセンサを取得する】
            KS_KinectSensor = (from sensorToCheck
                               in KinectSensor.KinectSensors
                               where sensorToCheck.Status == KinectStatus.Connected
                               select sensorToCheck).FirstOrDefault();
            //【検出結果の表示】
            if (KS_KinectSensor == null) //【Kinect接続が確認できなかった場合】
            {
                pictureBox2.Image = imageList1.Images[1];
                MessageBox.Show(
                   "接続可能なKINECTセンサが見当たりません。",
                   "KINECT for Windows Error Message",
                   MessageBoxButtons.OK
                );
                return;
            }
            else //【Kinect検出に成功】
            {
                pictureBox2.Image = imageList1.Images[0];
            }

            //***********************************************************
            //【Kinectの初期化処理】
            //***********************************************************
            //【KINECTの設定】ColorStream, DepthStream, SkeletonStreamを許可
            KS_KinectSensor.ColorStream.Enable(ColorImageFormat.RgbResolution640x480Fps30); //【ColorStream許可】
            KS_KinectSensor.DepthStream.Enable(DepthImageFormat.Resolution320x240Fps30);    //【DepthStream許可】
            KS_KinectSensor.SkeletonStream.Enable();                                        //【SkeltonStream許可】
            //【★★★デプス情報のPlayerIDが有効になるのはSkeletonStreamを有効にしている場合に限られるので注意★★★】
            //【KINECT動作開始】
            KS_KinectSensor.Start(); //【KINECTスタート】

            //【KINECTデータ取得イベントの設定とデータバッファの確保】
            KS_KinectSensor.ColorFrameReady += new EventHandler<ColorImageFrameReadyEventArgs>(ColorImageReady); //【準備完了でColorImageReadyがcallされる】
            KS_KinectSensor.DepthFrameReady += new EventHandler<DepthImageFrameReadyEventArgs>(DepthImageReady); //【準備完了でDepthImageReadyがcallされる】
            KS_KinectSensor.AllFramesReady += new EventHandler<AllFramesReadyEventArgs>(KinectAllFramesReady);
            KS_ColorImageDataXRGB = new byte[KS_KinectSensor.ColorStream.FramePixelDataLength]; //【カラー画像データ用】
            KS_ColorImageDataARGB = new byte[KS_KinectSensor.ColorStream.FramePixelDataLength]; //【カラー画像データ用】
            KS_ColorFrameWidth = KS_KinectSensor.ColorStream.FrameWidth;   //【ColorImageの横方向画素数】
            KS_ColorFrameHeight = KS_KinectSensor.ColorStream.FrameHeight; //【ColorImageの縦方向画素数】
            KS_bmp1 = new Bitmap(KS_ColorFrameWidth, KS_ColorFrameHeight, System.Drawing.Imaging.PixelFormat.Format32bppArgb); //【ColorImage表示用】
            KS_rect1 = new Rectangle(0, 0, KS_ColorFrameWidth, KS_ColorFrameHeight); //【ColorImage用枠】
            ClearColorImage(); //【カラー画像画面クリア】透明
            KS_DepthData = new short[KS_KinectSensor.DepthStream.FramePixelDataLength];              //【深度情報データ用】
            KS_DepthImageDataARGB = new byte[KS_KinectSensor.DepthStream.FramePixelDataLength * 4];  //【深度情報データ表示用】
            KS_DepthFrameWidth = KS_KinectSensor.DepthStream.FrameWidth;   //【DepthImageの横方向画素数】
            KS_DepthFrameHeight = KS_KinectSensor.DepthStream.FrameHeight; //【DepthImageの縦方向画素数】
            KS_bmp2 = new Bitmap(KS_DepthFrameWidth, KS_DepthFrameHeight, System.Drawing.Imaging.PixelFormat.Format32bppArgb); //【DepthImage表示用】
            KS_rect2 = new Rectangle(0, 0, KS_DepthFrameWidth, KS_DepthFrameHeight); //【DepthImage用枠】
            KS_SkeletonData = new Skeleton[KS_KinectSensor.SkeletonStream.FrameSkeletonArrayLength]; //【骨格情報用】
            KS_PlayerIndexBitmask = DepthImageFrame.PlayerIndexBitmask;
            KS_PlayerIndexBitmaskWidth = DepthImageFrame.PlayerIndexBitmaskWidth;
            KS_KinectSensor.DepthStream.Range = DepthRange.Default;    //【Depth測定モードをDefaultに設定】
            KS_CheckDistance();                                        //【設定結果を表示】Lmax, Lminを取得

            int Angle = 0;
            KS_KinectSensor.ElevationAngle = Angle; //【Kinectヘッドの仰角を設定】設定範囲は±27°
            textBox1.Text = Angle.ToString();

            ColorBar.CreateColorPalletForAbsolute(KS_Lmax + 1); //【カラーパレットを作成する】
            pictureBox3.Image = ColorBar.CreateColorScale(256, 10);    //【256×10画素サイズのカラーバーを作成して表示する】

            pictureBox4.SizeMode = PictureBoxSizeMode.StretchImage; //【デプス・イメージがpictureBox枠内にはめ込まれます】
            pictureBox1.SizeMode = PictureBoxSizeMode.StretchImage; //【キャプチャー画像がpictureBox枠内にはめ込まれます】
            pictureBox1.BackgroundImage = System.Drawing.Image.FromFile(BackGroundFileName[1]); //【Sky.bmpを表示する】
            NTL_SkeletonData = new SkeletonData();
            NTL_FIFO_Initialize(); //【FIFOをdefault段数で初期化】

            this.FormClosing += new FormClosingEventHandler(Form1_FormClosing);
            radioButton1.CheckedChanged += new EventHandler(DepthRangeChange);
            radioButton3.CheckedChanged += new EventHandler(BackgroundChange);
            checkBox1.CheckedChanged += new EventHandler(checkBox1_CheckedChanged);
        }

        //***************************************************************
        //【終了処理】
        //***************************************************************
        void Form1_FormClosing(object sender, FormClosingEventArgs e)
        {
            if (KS_KinectSensor != null)
            {
                KS_KinectSensor.Stop();    //【KINECTセンサを終了】★これを実行しないとレーザーが消えない
                KS_KinectSensor.Dispose(); //【KINECTセンサのオブジェクトを解放する】
            }
        }

        //***************************************************************
        //【距離モード変更】Defaultモード(800mm～4000mm) Nearモード(400m～3000mm)
        //***************************************************************
        void DepthRangeChange(object sender, EventArgs e)
        {
            if (radioButton1.Checked) { KS_KinectSensor.DepthStream.Range = DepthRange.Default; label8.Text = "Default"; }
            else { KS_KinectSensor.DepthStream.Range = DepthRange.Near; label8.Text = "Near"; }
            KS_CheckDistance();
        }

        //***************************************************************
        //【CheckDistance】KINECTセンサの検出可能距離現在値を調べる
        //***************************************************************
        private void KS_CheckDistance()
        {
            KS_Lmax = KS_KinectSensor.DepthStream.MaxDepth;
            KS_Lmin = KS_KinectSensor.DepthStream.MinDepth;
            label1.Text = "Min=" + KS_Lmin.ToString() + "mm";
            label2.Text = "Max=" + KS_Lmax.ToString() + "mm";
        }

        //***************************************************************
        //【背景画変更】
        //***************************************************************
        void BackgroundChange(object sender, EventArgs e)
        {
            if (radioButton3.Checked) { pictureBox1.BackgroundImage = System.Drawing.Image.FromFile(BackGroundFileName[0]); }
            else { pictureBox1.BackgroundImage = System.Drawing.Image.FromFile(BackGroundFileName[1]); }
        }

        //***************************************************************
        //【上下角変更ボタン】KINECTの上下角を変更。±27°の範囲内。時間を置かないとエラーが発生する
        //***************************************************************
        private void button2_Click(object sender, EventArgs e)
        {
            Int32 angle = int.Parse(textBox1.Text); //【設定角度を読み込む】±27°
            if (angle <= KS_KinectSensor.MaxElevationAngle && angle >= KS_KinectSensor.MinElevationAngle)
            {
                KS_KinectSensor.ElevationAngle = angle; //【上下角の変更】
            }
        }

        //***************************************************************
        //【カラー画像の左右反転指定フラグ】
        //***************************************************************
        private void checkBox1_CheckedChanged(object sender, EventArgs e)
        {
            if (checkBox1.Checked) { KS_flagFlip = true; } else { KS_flagFlip = false; }
        }

        //***************************************************************
        //【ColorImageの取得イベント】
        //***************************************************************
        private void ColorImageReady(object sender, ColorImageFrameReadyEventArgs e)
        {
            using (KS_ColorFrame = e.OpenColorImageFrame())
            {
                if (KS_ColorFrame != null)
                {
                    int j, k1, k2;
                    KS_ColorFrame.CopyPixelDataTo(KS_ColorImageDataXRGB); //【KINECTのColor画像を配列PixelDataにコピー】
                    if (KS_flagPlayer == false) return;
                    CopyPlayerBody(); //【デプス情報に基づいてプレーヤーを切り出す】
                    KS_data1 = KS_bmp1.LockBits(KS_rect1, ImageLockMode.WriteOnly, System.Drawing.Imaging.PixelFormat.Format32bppArgb); //【Lock】
                    Marshal.Copy(KS_ColorImageDataARGB, 0, KS_data1.Scan0, KS_ColorImageDataARGB.Length); //【UnmanagedMemoryのBitmapDataに配列PixelDataをコピー】
                    KS_bmp1.UnlockBits(KS_data1);           //【Lock解除】
                    if (KS_flagFlip) { KS_bmp1.RotateFlip(RotateFlipType.Rotate180FlipY); } //【左右反転】
                    pictureBox1.Image = KS_bmp1;            //【pictureBox1に画像表示】

                    Graphics g = Graphics.FromImage(pictureBox1.Image);
                    SolidBrush newbrush = new SolidBrush(Color.Red);
                    for (j = 0; j < NTL_SkeletonData.JointPosition.Length; j++)
                    {
                        g.FillEllipse(newbrush, (float)NTL_SkeletonData.JointPosition[j].X - 5f, (float)NTL_SkeletonData.JointPosition[j].Y - 5f, 10f, 10f);
                    }
                    for (j = 0; j < NTL_SkeletonData.Bone.Length; j++)
                    {
                        k1 = NTL_SkeletonData.BoneIndexList[2 * j];
                        k2 = NTL_SkeletonData.BoneIndexList[2 * j + 1];
                        g.DrawLine(new Pen(Color.Yellow, 3.0f),
                                   new PointF((float)NTL_SkeletonData.JointPosition[k1].X, (float)NTL_SkeletonData.JointPosition[k1].Y),
                                   new PointF((float)NTL_SkeletonData.JointPosition[k2].X, (float)NTL_SkeletonData.JointPosition[k2].Y) );
                    }
                 //   g.Dispose();

                    if (NTL_FIFOempty) return;
                 // g = Graphics.FromImage(pictureBox1.Image);
                    int i;
                    NTL_FIFOData a;
                    if (NTL_FIFOfull)
                    {
                        NTL_rPtr = NTL_wPtr;
                        newbrush = new SolidBrush(Color.Red);
                        for (i = 0; i < NTL_nFIFO; i++)
                        {
                            a = NTL_FIFO_Read_RH();
                            g.FillEllipse(newbrush, (float)a.Pc.X - 5f, (float)a.Pc.Y - 5f, 10f, 10f);
                        }
                        NTL_rPtr = NTL_wPtr;
                        newbrush = new SolidBrush(Color.Green);
                        for (i = 0; i < NTL_nFIFO; i++)
                        {
                            a = NTL_FIFO_Read_LH();
                            g.FillEllipse(newbrush, (float)a.Pc.X - 5f, (float)a.Pc.Y - 5f, 10f, 10f);
                        }
                    }
                    else
                    {
                        newbrush = new SolidBrush(Color.Red);
                        for (i = 0; i < NTL_wPtr; i++)
                        {
                            a = NTL_FIFO_Read_RH();
                            g.FillEllipse(newbrush, (float)a.Pc.X - 5f, (float)a.Pc.Y - 5f, 10f, 10f);
                        }
                        newbrush = new SolidBrush(Color.Green);
                        for (i = 0; i < NTL_wPtr; i++)
                        {
                            a = NTL_FIFO_Read_LH();
                            g.FillEllipse(newbrush, (float)a.Pc.X - 5f, (float)a.Pc.Y - 5f, 10f, 10f);
                        }
                    }
                    g.Dispose();

                    Validate();

                //  Application.DoEvents();
                }
                else
                {
                    //【ImageFrameがnullだったのは要求が間に合わなかったと考えられます】
                }
            }
        }

        //***************************************************************
        //【CopyPlayerBody】
        //***************************************************************
        private void CopyPlayerBody()
        {
            int i, j, k, L; float x, y;

            y = 0.0f;
            for (j = NTL_Pc1.Y; j < KS_ColorFrameHeight; j++)
            {
                x = 0.0f;
                for (i = NTL_Pc1.X; i < KS_ColorFrameWidth; i++)
                {
                    k = 4 * (i + j * KS_ColorFrameWidth);
                    if (((int)x < KS_DepthFrameWidth) && ((int)y < KS_DepthFrameHeight))
                    {
                        L = (int)x + (int)y * KS_DepthFrameWidth;
                        if ((KS_DepthData[L] & KS_PlayerIndexBitmask) != 0)
                        {
                            KS_ColorImageDataARGB[k] = KS_ColorImageDataXRGB[k];
                            KS_ColorImageDataARGB[k + 1] = KS_ColorImageDataXRGB[k + 1];
                            KS_ColorImageDataARGB[k + 2] = KS_ColorImageDataXRGB[k + 2];
                            KS_ColorImageDataARGB[k + 3] = 255; // KS_ColorImageDataXRGB[k + 3];
                        }
                        else
                        {
                            KS_ColorImageDataARGB[k] = 0;
                            KS_ColorImageDataARGB[k + 1] = 0;
                            KS_ColorImageDataARGB[k + 2] = 0;
                            KS_ColorImageDataARGB[k + 3] = 1;
                        }
                        x += NTL_cDx;
                    }
                }
                y += NTL_cDy;
            }
        }

        //***************************************************************
        //【ClearColorImage】カラー画像画面全体を透明にする
        //***************************************************************
        private void ClearColorImage()
        {
            for (int i = 0; i < KS_ColorImageDataARGB.Length; i += 4)
            {
                KS_ColorImageDataARGB[i] = 0;
                KS_ColorImageDataARGB[i + 1] = 0;
                KS_ColorImageDataARGB[i + 2] = 0;
                KS_ColorImageDataARGB[i + 3] = 1;
            }
        }

        //***************************************************************
        //【DepthImageの取得イベント】
        //***************************************************************
        private void DepthImageReady(object sender, DepthImageFrameReadyEventArgs e)
        {
            using (KS_DepthFrame = e.OpenDepthImageFrame())
            {
                if (KS_DepthFrame != null)
                {
                    //kinectSensor.DepthStreamに以下のフラグがある。
                    // .TooNearDepth【距離が近すぎる】, .TooFarDepth【距離が遠すぎる】, .UnknownDepth【距離が不明】
                    KS_DepthFrame.CopyPixelDataTo(KS_DepthData); //【KINECTのDepth情報を配列DepthDataにコピー】
                    KS_ConvertDepthToColor(); //【深度情報をカラーバーの色を使って表示形式をA8R8G8B8形式に変換】

                    KS_data2 = KS_bmp2.LockBits(KS_rect2, ImageLockMode.WriteOnly, System.Drawing.Imaging.PixelFormat.Format32bppArgb); //【Lockする】
                    Marshal.Copy(KS_DepthImageDataARGB, 0, KS_data2.Scan0, KS_DepthImageDataARGB.Length); //【UnmanagedMemoryのBitmapDataに配列depthDataをコピーする】
                    KS_bmp2.UnlockBits(KS_data2); //【Lock解除】
                    if (KS_flagFlip) { KS_bmp2.RotateFlip(RotateFlipType.Rotate180FlipY); } //【左右反転】
                    pictureBox4.Image = KS_bmp2;  //【pictureBox1に画像表示】

                    if (NTL_flagMappingInfo)
                    {
                        //【マッピング用】カラーイメージ⇒デプスイメージ
                        NTL_Pc1 = KS_DepthFrame.MapToColorImagePoint(0, 0, ColorImageFormat.RgbResolution640x480Fps30);
                        NTL_Pc2 = KS_DepthFrame.MapToColorImagePoint(319, 0, ColorImageFormat.RgbResolution640x480Fps30);
                        NTL_Pc3 = KS_DepthFrame.MapToColorImagePoint(0, 239, ColorImageFormat.RgbResolution640x480Fps30);
                        NTL_cDx = 319.0f / (NTL_Pc2.X - NTL_Pc1.X);
                        NTL_cDy = 239.0f / (NTL_Pc3.Y - NTL_Pc1.Y);
                        NTL_flagMappingInfo = false;
                    }
                }
                else
                {
                    //【DepthFrameがnullだったのは要求が間に合わなかったと考えられます】
                }
            }
        }

        //***************************************************************
        //【ConvertDepthToColor】16ビットのDepth値をカラーバーに対応する色に変更
        //***************************************************************
        private void KS_ConvertDepthToColor()
        {
            int i, j, d;  //【Nearモード】  400～3000mm ⇒ 0.0980769　2600色
            int f; //【人物消失処理用】
            int playerID; //【下位3bitはplayerID】【未使用】
            if (KS_flagPlayer == false)
            {
                j = 0;
                for (i = 0; i < KS_DepthData.Length; i++)
                {
                    d = KS_DepthData[i];                  //【距離情報】【上位13bitがDepth値】
                    playerID = d & KS_PlayerIndexBitmask; //【下位3bitはplayerID】【未使用】
                    if (playerID > 0) { KS_flagPlayer = true; }
                    d = ((d >> KS_PlayerIndexBitmaskWidth) & 0x1FFF);  //【realDepth】0mm～8191mm
                    if (d < KS_Lmin) { d = KS_Lmin; } //【近すぎる】
                    if (d > KS_Lmax) { d = KS_Lmax; } //【遠すぎる】
                    d = d * 4;
                    KS_DepthImageDataARGB[j] = ColorBar.NTL_ColorElement[d]; //【青】B
                    KS_DepthImageDataARGB[j + 1] = ColorBar.NTL_ColorElement[d + 1]; //【緑】G
                    KS_DepthImageDataARGB[j + 2] = ColorBar.NTL_ColorElement[d + 2]; //【赤】R
                    KS_DepthImageDataARGB[j + 3] = ColorBar.NTL_ColorElement[d + 3]; //【α】
                    j = j + 4;
                }
            }
            else
            {
                f = 0;
                j = 0;
                for (i = 0; i < KS_DepthData.Length; i++)
                {
                    d = KS_DepthData[i];                  //【距離情報】【上位13bitがDepth値】
                    playerID = d & KS_PlayerIndexBitmask; //【下位3bitはplayerID】【未使用】
                    if (playerID > 0)
                    {
                        f = 1;
                        d = ((d >> KS_PlayerIndexBitmaskWidth) & 0x1FFF);  //【realDepth】0mm～8191mm
                        if (d < KS_Lmin) { d = KS_Lmin; } //【近すぎる】
                        if (d > KS_Lmax) { d = KS_Lmax; } //【遠すぎる】
                        d = d * 4;
                        KS_DepthImageDataARGB[j] = ColorBar.NTL_ColorElement[d];         //【青】B
                        KS_DepthImageDataARGB[j + 1] = ColorBar.NTL_ColorElement[d + 1]; //【緑】G
                        KS_DepthImageDataARGB[j + 2] = ColorBar.NTL_ColorElement[d + 2]; //【赤】R
                        KS_DepthImageDataARGB[j + 3] = ColorBar.NTL_ColorElement[d + 3]; //【α】
                    }
                    else
                    {   //【背景画像を透過させるために透明にする】
                        KS_DepthImageDataARGB[j] = 0;     //【青】B
                        KS_DepthImageDataARGB[j + 1] = 0; //【緑】G
                        KS_DepthImageDataARGB[j + 2] = 0; //【赤】R
                        KS_DepthImageDataARGB[j + 3] = 1; //【α】
                    }
                    j = j + 4;
                }
                if (f == 0) { KS_flagPlayer = false; }
            }
        }

        //***************************************************************
        //【SkeletonFrameの取得イベント】SkeletonFrameは6名分。各々がJointsコレクションを持つ。
        //***************************************************************
        private void KinectAllFramesReady(object sender, AllFramesReadyEventArgs e)
        {
            using (KS_SkeletonFrame = e.OpenSkeletonFrame())
            {
                if (KS_SkeletonFrame != null)
                {
                    KS_SkeletonFrame.CopySkeletonDataTo(KS_SkeletonData); //【KINECTのSkeltonデータを配列SkeletonDataにコピー】
                    for (int i = 0; i < KS_SkeletonData.Length; i++)
                    {
                        if (KS_SkeletonData[i].TrackingState == SkeletonTrackingState.Tracked)
                        {
                            label9.Text = "検出したプレイヤー番号＝" + i.ToString();
                            NTL_SkeletonData.SetBody(KS_KinectSensor, KS_SkeletonData[i], KS_flagFlip);

                            if (KS_SkeletonData[i].Joints[JointType.HandRight].TrackingState == JointTrackingState.Tracked)
                            {
                                NTL_RightHand.Ps = KS_SkeletonData[i].Joints[JointType.HandRight].Position; //【右手】
                                if (KS_flagFlip) { NTL_RightHand.Ps.X = -(NTL_RightHand.Ps.X); }
                                NTL_RightHand.Pd = KS_KinectSensor.MapSkeletonPointToDepth(NTL_RightHand.Ps, DepthImageFormat.Resolution640x480Fps30);
                                NTL_RightHand.Pc = KS_KinectSensor.MapSkeletonPointToColor(NTL_RightHand.Ps, ColorImageFormat.RgbResolution640x480Fps30);
                                NTL_RightHand.ColorValue = Color.Red; //【本来はセンサ値に応じて色を変更する】
                                NTL_RightHand.SensorValue = 0;  //【センサの計測値】未使用
                            }
                            if (KS_SkeletonData[i].Joints[JointType.HandLeft].TrackingState == JointTrackingState.Tracked)
                            {
                                NTL_LeftHand.Ps = KS_SkeletonData[i].Joints[JointType.HandLeft].Position; //【右手】
                                if (KS_flagFlip) { NTL_LeftHand.Ps.X = -(NTL_LeftHand.Ps.X); }
                                NTL_LeftHand.Pd = KS_KinectSensor.MapSkeletonPointToDepth(NTL_LeftHand.Ps, DepthImageFormat.Resolution640x480Fps30);
                                NTL_LeftHand.Pc = KS_KinectSensor.MapSkeletonPointToColor(NTL_LeftHand.Ps, ColorImageFormat.RgbResolution640x480Fps30);
                                NTL_LeftHand.ColorValue = Color.Green; //【本来はセンサ値に応じて色を変更する】
                                NTL_LeftHand.SensorValue = 0;  //【センサの計測値】未使用
                            }
                            NTL_FIFO_Push(NTL_RightHand, NTL_LeftHand);

                            break;
                        }
                    }
                }
                else
                {
                    //【SkeletonFrameがnullだったのは要求が間に合わなかったと考えられます】
                }
            }
        }

        //***************************************************************
        //【拡張現実センサ用のFIFO】
        //***************************************************************

        public void NTL_FIFO_Initialize(int n)
        {
            NTL_wPtr = 0;
            NTL_rPtr = 0;
            NTL_nFIFO = n;
            NTL_FIFO_RH = new NTL_FIFOData[NTL_nFIFO];
            NTL_FIFO_LH = new NTL_FIFOData[NTL_nFIFO];
        }

        public void NTL_FIFO_Initialize()
        {
            NTL_wPtr = 0;
            NTL_rPtr = 0;
            NTL_nFIFO = NTL_default_nFIFO;
            NTL_FIFO_RH = new NTL_FIFOData[NTL_nFIFO];
            NTL_FIFO_LH = new NTL_FIFOData[NTL_nFIFO];
        }

        public void NTL_FIFO_Push(NTL_FIFOData RH, NTL_FIFOData LH)
        {
            NTL_FIFO_RH[NTL_wPtr] = RH;
            NTL_FIFO_LH[NTL_wPtr] = LH;
            NTL_wPtr++;
            NTL_FIFOempty = false;
            if (NTL_wPtr == NTL_nFIFO) { NTL_wPtr = 0; NTL_FIFOfull = true; }
        }

        public NTL_FIFOData NTL_FIFO_Read_RH()
        {
            NTL_FIFOData a = NTL_FIFO_RH[NTL_rPtr];
            NTL_rPtr++;
            if (NTL_rPtr == NTL_nFIFO) { NTL_rPtr = 0; }
            return a;
        }

        public NTL_FIFOData NTL_FIFO_Read_LH()
        {
            NTL_FIFOData a = NTL_FIFO_LH[NTL_rPtr];
            NTL_rPtr++;
            if (NTL_rPtr == NTL_nFIFO) { NTL_rPtr = 0; }
            return a;
        }


    }
}
